from flask_restful import fields


class ISODateTime(fields.DateTime):
    def __init__(self):
        super(ISODateTime, self).__init__(dt_format="iso8601")

    def format(self, value):
        line = super(ISODateTime, self).format(value)
        line = line[:26]
        return line + "Z"


class RFC822DateTime(fields.DateTime):
    def __init__(self):
        super(RFC822DateTime, self).__init__(dt_format="rfc822")



import functools
import inspect
from datetime import datetime
from decimal import Decimal

import flask_restful.fields
from flask_restful import abort
from flask_restful.reqparse import Argument
from flask_restful.reqparse import RequestParser
from flask_restful_swagger import swagger
from six import add_metaclass
from six import iteritems
from six import text_type
from werkzeug.datastructures import FileStorage
from werkzeug.datastructures import MultiDict

from flask_kits.restful import post_parameter

LOC_JSON = 'json'
LOC_VALUES = 'values'
LOC_HEADERS = 'headers'
LOC_ARGS = 'args'
LOC_FILES = 'files'
LOC_FORM = 'form'
LOC_COOKIES = 'cookies'

MAPPINGS = {
    int: flask_restful.fields.Integer,
    float: flask_restful.fields.Float,
    text_type: flask_restful.fields.String,
    datetime: flask_restful.fields.DateTime,
    Decimal: flask_restful.fields.String,
    bool: flask_restful.fields.Boolean
}
SWAGGER_PARAMS = {
    LOC_HEADERS: 'header',
    LOC_JSON: 'body',
    LOC_FILES: 'body',
    LOC_ARGS: 'query'
}
SWAGGER_DATA = {
    FileStorage: 'file'
}


# optimize this code
def get_field_type(class_type):
    return MAPPINGS.get(class_type)


def swagger_param(param_type):
    """
    mapping flask_restful.reqparse.RequestParser to swagger 
    :param str param_type: 
    :rtype: str
    """
    param_type = param_type.lower()
    return SWAGGER_PARAMS.get(param_type, 'body')


def swagger_data(data_type):
    """
    
    :param data_type: 
    :return: 
    """
    return SWAGGER_DATA.get(data_type) or str(data_type)


class EntityDeclarative(type):
    def __new__(cls, class_name, bases, attributes):
        if class_name == 'EntityBase':
            return type.__new__(cls, class_name, bases, attributes)

        parser = RequestParser()
        fields = [(name, field) for name, field in iteritems(attributes) if isinstance(field, Field)]

        field_names = set()
        resource_fields = dict()
        extra_fields = []
        for key, field in fields:
            if inspect.isclass(field.type) and issubclass(field.type, EntityBase):
                resource_fields[key] = flask_restful.fields.Nested(field.type.resource_fields)
                field.type = field.type.parse
                field.location = 'json'
            elif field.location == 'json':
                resource_fields[key] = get_field_type(field.type)
            else:
                extra_fields.append(field)
            parser.add_argument(field)
            field_names.add(key)
            del attributes[key]
        attributes['entity_parser'] = parser
        attributes['entity_fields'] = field_names
        attributes['resource_fields'] = resource_fields
        attributes['extra_fields'] = extra_fields

        schema = type.__new__(cls, class_name, bases, attributes)
        # support swagger
        swagger.add_model(schema)
        return schema


# TODO(benjamin): optimize
class WrappedDict(dict):
    def __init__(self, source):
        super(WrappedDict, self).__init__(**source)

    def json(self):
        return MultiDict(self)

    def values(self):
        return None


@add_metaclass(EntityDeclarative)
class EntityBase(dict):
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self[name] = value

    @classmethod
    def parse(cls, req=None):
        """

        :param req: 
        :rtype: EntityBase
        """
        if isinstance(req, dict):
            req = WrappedDict(req)
        instance = cls()  # type: EntityBase
        args = cls.entity_parser.parse_args(req)
        for field in cls.entity_fields:
            setattr(instance, field, args[field])
        success = instance.validate()
        if isinstance(success, ValueError):
            abort(400, message={'error': text_type(success)})
        return instance

    @classmethod
    def parameter(cls, f):
        """
        :param EntityBase cls:
        :param f:
        """

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            entity = cls.parse()
            kwargs['entity'] = entity
            return f(*args, **kwargs)

        # Support swagger document
        if '__swagger_attr' in f.__dict__:
            attr = wrapped.__dict__['__swagger_attr'] = f.__dict__['__swagger_attr']
            attr['consumes'] = ["multipart/form-data", 'application/json']
            params = attr.get('parameters', [])
            if cls.resource_fields:
                params.append(post_parameter(cls))
            for field in cls.extra_fields:
                params.append({
                    'name': field.name,
                    'description': field.name,
                    'required': field.required,
                    'dataType': swagger_data(field.type),
                    'paramType': swagger_param(field.location)
                })
        return wrapped

    def validate(self):
        """
        :rtype: (bool|ValueError)
        """
        return True


class Field(Argument):
    def __init__(self, name, *args, **kwargs):
        kwargs.setdefault('type', text_type)
        self.validators = set()
        kwargs.setdefault('location', 'json')
        if 'validators' in kwargs:
            self.validators = kwargs.pop('validators')
        super(Field, self).__init__(name, *args, **kwargs)

    def parse(self, request, bundle_errors=False):
        value, found = super(Field, self).parse(request, bundle_errors)
        if not isinstance(value, ValueError) and self.validators:
            for validator in self.validators:
                success = validator(value)
                if isinstance(success, ValueError):
                    found = {self.name: text_type(success)}
                    value = ValueError()
                    break
        return value, found



# -:- coding:utf8 -:-
from flask_kits.exceptions import InvalidFormatError


class Pagination(object):
    """
    Example
    >>> from flask_kits.restful import entity.Pagination
    >>> class ItemsPagination(Pagination):
    >>>     limit = 20
    >>>     ORDERS = ('ID', 'NAME')
    >>> class ItemEntity(EntityBase):
    >>>     Range = Field("Range", type=ItemsPagination, location=LOC_HEADERS, default=DirectoryPagination.default)
    """
    DIRECTIONS = ('ASC', 'DESC')
    ORDERS = ('ID',)
    start = 0
    limit = 100
    order_by = 'ID'
    order = 'ASC'

    def __init__(self, line=None):
        self._parse(line)

    def _parse(self, line):
        line = line or ''
        segments = [x.strip() for x in line.split(';', 1) if x.strip()]
        if not segments:
            return

        try:
            segment, segments = segments[0], segments[1:]
            order_by, start_and_end = segment.split(' ', 1)
            if order_by not in self.ORDERS:
                raise InvalidFormatError()
            self.order_by = order_by

            start, end = start_and_end.split('..')
            start = int(start or str(self.start))
            self.start = max(start, 0)
            # end = int(end or str(self.start + self.limit))
            # if not (0 < start < end):
            #     raise InvalidFormatError()
            # self.start = start
            # self.limit = end - start

            if segments:
                segment, = segments
                options = dict([x.split('=') for x in segment.split(',') if x.strip()])
                if 'order' in options:
                    order = options.get('order', self.order).upper()
                    if order not in self.DIRECTIONS:
                        order = Pagination.order
                    self.order = order

                if 'max' in options:
                    size = int(options.get('max'))
                    self.limit = min(max(size, 0), 1000)

        except InvalidFormatError as e:
            raise e
        except Exception:
            raise InvalidFormatError()

    def status_and_headers(self, rowcount, total_count=None):
        code = 206 if rowcount == self.limit else 200
        h = {'Accept-Range': ','.join(self.ORDERS)}
        if code == 206:
            total = '*' if total_count is None else total_count
            content_range = '{order_by} {start}..{end}/{total}'.format(
                order_by=self.order_by,
                start=self.start,
                end=self.start + rowcount,
                total=total)

            start = self.start + self.limit
            end = start + self.limit
            next_range = '{order_by} {start}..{end};order={order},max={limit}'.format(
                order=self.order,
                start=start,
                end=end,
                order_by=self.order_by,
                limit=self.limit
            )
            h['Next-Range'] = next_range
            h['Content-Range'] = content_range

        return code, h

    @classmethod
    def default(cls):
        return cls()

    @property
    def order_string(self):
        return "{0} {1}".format(self.order_by, self.order)



# -:- coding:utf8 -:-


class KitsError(Exception):
    pass


class RestfulError(KitsError):
    pass


class InvalidFormatError(RestfulError):
    def __init__(self):
        super(InvalidFormatError, self).__init__("range must be like: 'id ..;order=asc,max=20;'")



from .common import EntityBase
from .common import Field
from .common import LOC_ARGS
from .common import LOC_COOKIES
from .common import LOC_FILES
from .common import LOC_FORM
from .common import LOC_HEADERS
from .common import LOC_JSON
from .pagination import Pagination
from .response import Response
from .validator import LetterValidator
from .validator import MaxLengthValidator
from .validator import MinLengthValidator
from .validator import MoreValidator
from .validator import PrecisionValidator
from .validator import Validator



import re
from decimal import Decimal

import six


class Validator(object):
    help = ""

    def __call__(self, value):
        """
        validate param
        :param value: 
        :rtype: bool 
        """
        return self.validate(value)

    def validate(self, value):
        """
        :rtype: (None|ValueError)
        """
        raise NotImplementedError()

    def handle_error(self, *args):
        return ValueError(self.help.format(*args))


class CompareValidator(Validator):
    def __init__(self, threshold):
        self.threshold = threshold

    def validate(self, value):
        if value is None or self.illegal(value):
            return self.handle_error(self.threshold)

    def illegal(self, value):
        raise NotImplementedError()


class LetterValidator(CompareValidator):
    help = "Must be less than {0}"

    def illegal(self, value):
        return value > self.threshold


class MoreValidator(CompareValidator):
    help = "Must be more than {0}"

    def illegal(self, value):
        return value < self.threshold


class MinLengthValidator(CompareValidator):
    help = "String length must be more than {0}"

    def illegal(self, value):
        return len(value) < self.threshold


class MaxLengthValidator(CompareValidator):
    help = "String length must be less than {0}"

    def illegal(self, value):
        return len(value) > self.threshold


class PrecisionValidator(CompareValidator):
    help = "Must be less than {0} precision bit"

    def illegal(self, value):
        """
        :param Decimal value:
        """
        return value.quantize(Decimal('1.' + (self.threshold * '0'))) != value


class PasswordValidator(Validator):
    def validate(self, value):
        return True


class RegexValidator(Validator):
    help = "String length must be match the regex {0}"

    def __init__(self, line):
        self.conditions = []
        if isinstance(line, six.string_types):
            self.conditions = [re.compile(line, re.MULTILINE)]
        elif isinstance(line, (list, tuple)):
            self.conditions = [re.compile(x, re.MULTILINE) for x in line if x]
        else:
            self.conditions = [line]

    def validate(self, value):
        for condition in self.conditions:
            if not re.match(condition, value):
                return self.handle_error(condition.pattern)
        return True



import unittest

from flask_kits.restful.entity.validator import RegexValidator


class HelloWorldTestCase(unittest.TestCase):
    def test_regex_validator(self):
        validator = RegexValidator("^[A-Za-z0-9\-_]+$")
        self.assertTrue(validator.validate("abcd"))
        result = validator.validate("a/bcd")
        self.assertTrue(isinstance(result, ValueError))
